<?php

namespace NCSU\GPA;

final class GPA
{
    /**
     * Number of decimal places to present.
     * @see http://policies.ncsu.edu/regulation/reg-02-50-03
     * @var integer
     */
    const DECIMAL_PLACES = 3;

    /**
     * @var float the internal representation of grade point average.
     */
    private $gpa;

    private function __construct($gpa)
    {
        Assertion::float($gpa);
        $this->gpa = $gpa;
    }

    /**
     * Creates a grade point value object from a valid native representation.
     *
     * @param  numeric $gpa
     * @return GPA
     */
    public static function fromNative($gpa)
    {
        if (is_numeric($gpa)) {
            $gpa = (float) $gpa;
        }

        return new self($gpa);
    }

    /**
     * Creates a grade point value object from a valid letter grade.
     *
     * @param  string $grade
     * @return GPA
     */
    public static function fromGrade($grade)
    {
        $gradeToGpaValue = array(
            'A+' => 4.333,
            'A'  => 4.000,
            'A-' => 3.667,
            'B+' => 3.333,
            'B'  => 3.000,
            'B-' => 2.667,
            'C+' => 2.333,
            'C'  => 2.000,
            'C-' => 1.667,
            'D+' => 1.333,
            'D'  => 1.000,
            'D-' => 0.667,
            'F'  => 0.000,
        );

        Assertion::choice($grade, array_keys($gradeToGpaValue));

        return new self($gradeToGpaValue[$grade]);
    }

    /**
     * Returns a truncated representation of grade point value.
     * @return float
     */
    public function points()
    {
        $rounded = round($this->gpa, self::DECIMAL_PLACES, PHP_ROUND_HALF_UP);

        if (strlen($rounded) !== strlen($this->gpa)) {
            trigger_error('', E_USER_WARNING);
        }

        return $rounded;
    }

    /**
     * Converts a grade point value to ASCII grade letter.
     * @return string
     */
    public function grade()
    {
        if ($this->gpa >= 4.333) {
            return 'A+';
        }

        if ($this->gpaInRange(4.000, 4.332)) {
            return 'A';
        }

        if ($this->gpaInRange(3.667, 3.999)) {
            return 'A-';
        }

        if ($this->gpaInRange(3.333, 3.666)) {
            return 'B+';
        }

        if ($this->gpaInRange(3.000, 3.332)) {
            return 'B';
        }

        if ($this->gpaInRange(2.667, 2.999)) {
            return 'B-';
        }

        if ($this->gpaInRange(2.333, 2.666)) {
            return 'C+';
        }

        if ($this->gpaInRange(2.000, 2.332)) {
            return 'C';
        }

        if ($this->gpaInRange(1.667, 1.999)) {
            return 'C-';
        }

        if ($this->gpaInRange(1.333, 1.666)) {
            return 'D+';
        }

        if ($this->gpaInRange(1.000, 1.332)) {
            return 'D';
        }

        if ($this->gpaInRange(0.667, 0.999)) {
            return 'D-';
        }

        if ($this->gpaInRange(0.000, 0.666)) {
            return 'F';
        }
    }

    /**
     * @param float $low
     * @param float $high
     *
     * @return bool true if $gpa is in range [$low, $high].
     */
    private function gpaInRange($low, $high)
    {
        return $this->gpa >= $low && $this->gpa <= $high;
    }

    /**
     * @param GPA $other
     *
     * @return bool true if called GPA is greater than or equal to $other GPA.
     */
    public function gte(GPA $other)
    {
        return $this->gpa >= $other->gpa;
    }

    /**
     * @param GPA $other
     *
     * @return bool true if called GPA is greater than $other GPA.
     */
    public function gt(GPA $other)
    {
        return $this->gpa > $other->gpa;
    }

    /**
     * @param GPA $other
     *
     * @return bool true if called GPA is less than or equal to $other GPA.
     */
    public function lte(GPA $other)
    {
        return $this->gpa <= $other->gpa;
    }

    /**
     * @param GPA $other
     *
     * @return bool true if called GPA is less than $other GPA.
     */
    public function lt(GPA $other)
    {
        return $this->gpa < $other->gpa;
    }

    /**
     * @return bool true if GPA value meets "C-wall" requirements, false otherwise.
     */
    public function meetsCWall()
    {
        return $this->gpa >= 1.667;
    }
}
